#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <asm/delay.h>
#include "gpib.h"

/* define which parms can be altered */

#define is_digit(c)     ((c) >= '0' && (c) <= '9')    
#define GPIB_MOD_NAME	"gpib"
static int gpib_major=GPIB_MAJOR;
static int gpib_base_address=GPIB_BASE_ADDRESS;
static int gpib_device_number=GPIB_DEVICE_NUMBER;

MODULE_PARM(gpib_major,"i");
MODULE_PARM(gpib_minor,"i");
MODULE_PARM(gpib_base_address,"i");
MODULE_PARM(gpib_device_number,"i");
/*MODULE_PARM(gpib_irq,"i");*//*not sure if card does interrupts!*/
/*MODULE_AUTHOR("Matt Clark, Sept 2001");*/

static int gpib_ioctl(struct inode *, struct file *, unsigned int , unsigned long);
static ssize_t gpib_read(struct file *, char *, size_t, loff_t *);
static ssize_t gpib_write(struct file *, const char *, size_t, loff_t *);
static int gpib_open(struct inode *, struct file *);
static int gpib_release(struct inode *, struct file *);

/* #define TESTING	0*/

/* usleep=udelay */


struct GPIB_bus_status_struct{
	int	status;
	int 	device;
	};

static struct GPIB_bus_status_struct GPIB_bus_status;

struct file_operations gpib_fops = {
	read:		gpib_read,
	write:		gpib_write,
	ioctl:		gpib_ioctl,
	open:		gpib_open,
	release:	gpib_release,
	};

static int gpib_ioctl(struct inode * inode,struct file *file, unsigned int cmd, unsigned long arg){
gpd_t *gb;
	if (!(file->private_data))  return(0);
	gb=(gpd_t *)file->private_data;
	switch(cmd){
		case GPIB_HARDRESET:
			/* resets the usage count to 1
			   for use after a module crash
			   make sure nothing else has opened
			   the module */
			printk("gpib: ioctl: HARDRESET: attempting mod_use=1 ");
			/* pull it to 0 */
			while(MOD_IN_USE)MOD_DEC_USE_COUNT;
			MOD_INC_USE_COUNT;
			/* inc once for this open */
			printk("gpib: Done\n");
			break;
		default:
			printk("gpib: ioctl: Unknown ioctl cmd %#x\n",(int)cmd);
			}
	return 0;

	}


int init_module(void) {
int res,i,err;
	res=register_chrdev(gpib_major,GPIB_MOD_NAME,&gpib_fops);
	if (res<0) return(res);
	gpib_major=res;
	GPIB_bus_status.status=GPIB_IDLE;
	GPIB_bus_status.device=0;
		
#ifndef TESTING
	printk("gpib: init_module: about to initboard at baseaddress %d and GPIB bus number %d\n",gpib_base_address,gpib_device_number);
	res=GPIB_initboard(gpib_base_address,gpib_device_number);
	if(res<0) return res;
/* now register the ioports */
	printk("gpib:init:checking and registering ioports\n");
	for(i=0;i<8;i++){
		if((err=check_region(gpib_base_address+(i*GPIB_REG_STEP),1))<0){
			printk("gpib:init:unable to register ioport\n");
			return err;
			}
		}
	for(i=0;i<8;i++){
		request_region(gpib_base_address+(i*GPIB_REG_STEP),1,GPIB_MOD_NAME);
		}

	printk("gpib: init_module: board init'd ok\n");
#endif
#ifdef TESTING
	
#endif
	printk("gpib dev: init_module: Matt Clark aug 2001- device %d\n",gpib_major);
	return(0);
	}     

void cleanup_module(void){
int res,i;
	printk("gpib dev: cleanup_module: Matt Clark aug 2001- device %d\n",gpib_major);
	for(i=0;i<8;i++){
		release_region(gpib_base_address+(i*GPIB_REG_STEP),1);
		}    
	res=unregister_chrdev(gpib_major,GPIB_MOD_NAME);
	}    


/* read is complicated- there is no way of knowing in advance how big the message
   will be- this is determined by the device itself and while the user can query
   the device we can't (since we don't know what it will be).  The best we can do 
   is implement a system of EOF after every complete message and use a transfer
   buffer at least as big as the transfer buffer used by read

   To try to get sensible behaviour a partial read will lock the bus to this device
   and continue to read bytes of subsequent reads.

   A complete read results in EOF on the next read.

   A read of 0 bytes or EOF unlocks the bus
*/
 
static ssize_t gpib_read(struct file *file, char * buf,size_t count, loff_t *ppos) {
int dev,rem;
gpd_t *gb;
	if (!(file->private_data))  return(0);
	gb=(gpd_t *)file->private_data;
	dev=gb->device_number;
	printk("Gpib: in read\n");
	if(gb->controller){
		/* we can ignore the bus busy lock here especially as we don't do anything! */
		printk("gpib:read:reading controller %d\n",dev);
		return 0;
		}
	else{

		/* this is complicated and needs 2 locks.
		 * lock 1 (bus.status.xx) indicates if the bus is in use by another
		 *	device and return BUSY
		 * lock 2 (gb->next) indicates if we should
		 *	(a) return EOF rather than try the bus again (because it is end of message).
		 *	(b) continue reading bytes because the is remaining message pending
		 *	(c) the device is idle so we should query it again
		 */
		
		if(gb->next==GPIB_IO_ERROR){
			printk("gpib:read:Re-reading an IO ERROR\n");
			GPIB_bus_status.status=GPIB_IDLE;
			GPIB_bus_status.device=GPIB_NO_DEV;
			return -EIO;
			}
		/* check we are idle or in this read */
		if((GPIB_bus_status.status==GPIB_IDLE)||((GPIB_bus_status.status==GPIB_WAIT_DEV_READ)&&(GPIB_bus_status.device==dev))){
			if(gb->next==GPIB_NEXT_EOF){
				printk("gpib:read: returning EOF and to state idle\n");
				gb->next=GPIB_IDLE;
				GPIB_bus_status.status=GPIB_IDLE;
				GPIB_bus_status.device=dev;
				return 0;
				}
			printk("gpib:read: going to try to read device %d\n",dev);
			GPIB_bus_status.status=GPIB_WAIT_DEV_READ;
			GPIB_bus_status.device=dev;
			printk("gpib:read:reading device %d\n",dev);
			/* (re)allocate the transfer buffer */
			if(count > gb->buff_size){
				/* reallocate buffer !*/
				printk("gpib:read:device %d Growing buffer from %d to %d\n",dev,gb->buff_size,(int)count);
				if(buf!=NULL){
					kfree(gb->buffer);
					}	
				gb->buffer=kmalloc(count,GFP_KERNEL);
				if(gb->buffer==NULL){
					printk("gpib:read:device %d:unable to get %d bytes for buffer\n",dev,count);
					gb->buff_size=0;
					return -ENOMEM;
					}
				gb->buff_size=count;
				}

			/* now read the data from the gpib */

			/* if this is a new message set up else we have already done this */
			if(gb->next!=GPIB_NEXT_REMAINING_BYTES){
				printk("gpib:read:preparing to read bus for device %d\n",dev);
				/* first GPIB-config_to_receive */
				/* set set the board to unlisten.... */
				if(GPIB_sendcommand(GPIB_UNL_CMD)==GPIB_ERROR) return -EIO;
				/* set the board to listen */
				if(GPIB_sendcommand(GPIB_MLA_CMD+GPIB_board_gpib_address)==GPIB_ERROR) return -EIO;
				/* set the device to talk */
				if(GPIB_sendcommand(GPIB_MTA_CMD+dev)==GPIB_ERROR) return -EIO;
				if (GPIB_standby()!=GPIB_OK) return -EIO;    
				}

			//printk("gpib:read:attempting to read %d bytes from bus\n",count);
			/* now you can read the data */
			rem=GPIB_readdata(gb->buffer,count);
			//printk("gpib:read:readdata returned %d\n",rem);
			if(rem==GPIB_ERROR_TIMEOUT){
				printk("gpib:read: IO error on device %d- is it present / on ?\n",dev);
				printk("gpib:reseting the bus in case the Kiethley is connected\n");
				GPIB_initboard(gpib_base_address,gpib_device_number);
				gb->next=GPIB_IO_ERROR;
				GPIB_bus_status.status=GPIB_IDLE;
				GPIB_bus_status.device=GPIB_NO_DEV;
				return -EIO;
				}
			if(rem!=GPIB_ERROR_REMAINING_BYTES){
				/* this means a complete message has now been received */
				/* set the flags to get an EOF next */
				gb->next=GPIB_NEXT_EOF;
				}
			else{	/* this error means we read count bytes and
				   there were still bytes on the bus- an incomplete
				   read- the bus cannot be used by anyone else unless
				   this is cleared- matt Sept 2001 */
				gb->next=GPIB_NEXT_REMAINING_BYTES;
				rem=count;
				}	
			//printk("gpib:read:attempting to transfer %d bytes \n",rem);
			if(rem==0){
				//printk("there was nowt there so slip back to idle\n");
				/* there was nowt there so slip back to idle*/
				gb->next=GPIB_IDLE;
				GPIB_bus_status.status=GPIB_IDLE;
				GPIB_bus_status.device=GPIB_NO_DEV;
				return 0;
				}
			/* pass that data to user space */
			if(copy_to_user(buf,gb->buffer,rem)!=0){
				printk("gpib:read: copy_to_user: ERROR\n");
				}
			return rem;
			}
		else{
			//printk("we are busy doing something else\n");
			/* we are busy doing something else */
			return -EBUSY;
			}
		}
	/* never get here */
	}
 
/* write is simple cf read- we know how big the transfer is from the start so all we have to
   do is grab the data and squirt it over the bus */

static ssize_t gpib_write(struct file *file,const char * buf,size_t count, loff_t *ppos) {
gpd_t *gb;
int dev;
	if (!(file->private_data))  return(0);
	gb=(gpd_t *)file->private_data;
	dev=gb->device_number;
	if(gb->controller==TRUE){
		/* ignore the bus status here */
		//printk("gpib:write:writing controller %d\n",dev);
		return 0;
		}
	else{
                if(gb->next==GPIB_IO_ERROR){
                        printk("gpib:read:Re-reading an IO ERROR\n");
			GPIB_bus_status.status=GPIB_IDLE;
			GPIB_bus_status.device=GPIB_NO_DEV;
                        return -EIO;
                        }   
		//printk("gpib:write:writing device %d\n",dev);
		/* check we are not locked out by someone else */
		if((GPIB_bus_status.status==GPIB_IDLE)||((GPIB_bus_status.status==GPIB_WAIT_DEV_WRITE)&&(GPIB_bus_status.device==dev))){
		/* the second clause so always be ok as read should always be complete - matt sept 2001 */
			/* set the current status- shouldn't be require but for the sake of completeness */
			GPIB_bus_status.status=GPIB_WAIT_DEV_WRITE;
			GPIB_bus_status.device=dev;
			
			/* (re)allocate the transfer buffer */
			if(count > gb->buff_size){
				/* reallocate buffer !*/
				//printk("gpib:read:device %d Growing buffer from %d to %d\n",dev,gb->buff_size,(int)count);
				if(buf!=NULL){
					kfree(gb->buffer);
					}
				gb->buffer=kmalloc(count,GFP_KERNEL);
				if(gb->buffer==NULL){
					printk("gpib:read:device %d:unable to get %d bytes for buffer\n",dev,count);
					gb->buff_size=0;
					return -ENOMEM;
					}
				gb->buff_size=count;
				}   	
			/* now we have a buffer big enough get the data */
			if(copy_from_user(gb->buffer,buf,count)!=0){
				printk("gpib:read: copy_from_user: ERROR\n");
				}
			/* now we have the data squirt it down the bus */

			/* must select the dev here */
			/* set the board to talk*/
			if(GPIB_sendcommand(GPIB_MTA_CMD+GPIB_board_gpib_address)==GPIB_ERROR) return -EIO;
			/* set unlisten....*/
			if(GPIB_sendcommand(GPIB_UNL_CMD)==GPIB_ERROR) return -EIO;
			/* set the device to listen*/
			if(GPIB_sendcommand(GPIB_MLA_CMD+dev)==GPIB_ERROR) return -EIO;  
			if(GPIB_writedata(gb->buffer,count)==GPIB_ERROR_TIMEOUT){
				printk("gpib:write: IO error on device %d\n",dev);
				gb->next=GPIB_IO_ERROR;
				GPIB_bus_status.status=GPIB_IDLE;
				GPIB_bus_status.device=GPIB_NO_DEV;
				return -EIO;
				}
			GPIB_bus_status.status=GPIB_IDLE;
			GPIB_bus_status.device=GPIB_NO_DEV;
			gb->next=GPIB_IDLE;
			return count;	
			}
		else{
			/* the bus was locked by someone else */
			return -EBUSY;
			}
		}	
	/* never get here */
	}
 
/* this driver offers two devices gpib and gpib_text */
/* The bin version converts on read() and returns floats */
/* The text version converts on open() and returns text  */

static int gpib_open(struct inode *inode, struct file *file) {

gpd_t *gb;
int i;
	/* Should MOD_INC.. be here of after the possibility of returning with an error?*/
	MOD_INC_USE_COUNT;

	if ((file->private_data=kmalloc(sizeof(gpd_t),GFP_KERNEL))==NULL)
		return(-1);
	gb=(gpd_t *)file->private_data;
	gb->minor=MINOR(inode->i_rdev);
	if(gb->minor&GPIB_CONTROLLER_MASK){
		gb->controller=TRUE;
		gb->device_number=(gb->minor&GPIB_BOARD_MASK)>>5;
		printk("gpib:open:as controller on board %d\n",gb->device_number);
		}
	else{
		gb->controller=FALSE;
		gb->device_number=gb->minor;
		gb->next=GPIB_IDLE;
		gb->buff_size=0;
		gb->buffer=NULL;
		printk("gpib:open:as device %d\n",gb->device_number);
		}
	return 0;
	}  
 
static int gpib_release(struct inode *inode, struct file *file) {
gpd_t *gb;
	MOD_DEC_USE_COUNT;
	/* make sure that ALL memory alloc'd is freed matt Sept 2001*/
	if (file->private_data) {
        	gb=(gpd_t *)file->private_data; 
		if(gb->buffer!=NULL){
			kfree(gb->buffer);
			gb=NULL;
			}
		kfree(file->private_data);
		file->private_data=NULL;
		}
	return(0);
	}
      
